
#include "map"
#include "vector"

//Version 1: 3.8s   (6.12s)
//Version 2: 0.92s  (1.35s)

#define VERSION2 1

#ifdef VERSION2
//----------------
typedef int FamilyId;
struct EntitySystem;

struct Component {
};

struct Entity {
   static EntitySystem *entitySystem;
   Entity();
   template<typename Type> Type *getAs();
   std::map<FamilyId, Component*> mComponents;
};
EntitySystem *Entity::entitySystem = 0;

struct EntitySystem {
   EntitySystem() {
      Entity::entitySystem = this;
   }
   template<typename T> T *getComponent(Entity *e) {
      return (T*)e->mComponents[T::familyId];
   }
   template<typename T> void getEntities(std::vector<Entity*> &result) {
      auto iterPair = mComponentStore.equal_range(T::familyId);
      for(auto iter = iterPair.first; iter != iterPair.second; ++iter) {
         result.push_back(iter->second);
      }
   }
   template<typename T> void addComponent(Entity *e, T* comp) {
      mComponentStore.insert(std::pair<FamilyId, Entity*>(T::familyId, e));
      e->mComponents.insert(std::pair<FamilyId, Component*>(T::familyId, comp));
   }

   void deleteEntity(Entity *e) {
      //Scan Entity Components and get familyIds, then I know what to delete in the ComponentStore
      delete e;
   }
protected:
   std::multimap<FamilyId, Entity*> mComponentStore;
};

Entity::Entity() {
}

template<typename Type> Type *Entity::getAs() {
   return entitySystem->getComponent<Type>(this);
}

#endif

#ifdef VERSION1
//----------------
typedef int FamilyId;
struct EntitySystem;

struct Component {
};

struct Entity {
   static EntitySystem *entitySystem;
   Entity();
   template<typename Type> Type *getAs();
};
EntitySystem *Entity::entitySystem = 0;

struct EntitySystem {
   EntitySystem() {
      Entity::entitySystem = this;
   }
   template<typename T> T *getComponent(Entity *e) {
      std::map<Entity*, Component*> &store = mComponentStore[T::familyId];
      return (T*)store[e];
   }
   template<typename T> void getEntities(std::vector<Entity*> &result) {
      auto iter = mComponentStore[T::familyId].begin();
      auto endIter = mComponentStore[T::familyId].end();
      while(iter != endIter) {
         result.push_back(iter->first);
         iter++;
      }
   }
   template<typename T> void addComponent(Entity *e, T* comp) {
      mComponentStore[T::familyId].insert(std::pair<Entity*, Component*>(e, comp));
   }
   void deleteEntity(Entity *e) {
      auto familyIter = mComponentStore.begin();
      auto familyIterEnd = mComponentStore.end();
      while(familyIter != familyIterEnd) {
         auto compIter = (*familyIter).second.find(e);
         if(compIter != (*familyIter).second.end()) {
            delete (*compIter).second;          //delete component
            (*familyIter).second.erase(e);      //and remove from list
         }
         familyIter++;
      }
      delete e;
   }

protected:
   std::map<FamilyId, std::map<Entity*, Component*> > mComponentStore;
};

Entity::Entity() {
}

template<typename Type> Type *Entity::getAs() {
   return entitySystem->getComponent<Type>(this);
}

#endif
